#------------------------------------------------------------------------
#
#   IBM Software
#   Licensed Material - Property of IBM
#   (C) Copyright International Business Machines Corp. 2011, 2017
#   All Rights Reserved.
#
#   U.S. Government Users Restricted Rights - Use, duplication or
#   disclosure restricted by GSA ADP Schedule Contract with
#   IBM Corp.
#
#------------------------------------------------------------------------
# Syscall trace
#------------------------------------------------------------------------

namespace eval syscall_trace {
    variable version             "1.1"
    variable verbose              0
    variable max_syscall_number   256
}

package provide syscall_trace $syscall_trace::version

# Note: The mapping of syscall numbers to functions can be found in the Linux
# source in arch/powerpc/include/asm/unistd.h.
# Function prototypes for syscall functions can be found in the Linux
# source in include/linux/syscalls.h.

proc syscall_trace::get_str { sim cpu ea } {
    if {[catch {
        set pa [$sim cpu $cpu dtranslate $ea]
        set str "\"[$sim memory display $pa STRING]\""
    } err]} { 
        set str [format "0x%x" $ea]
    }
    return $str
}

proc syscall_trace::syscall_trigger { argstring } {
    if {[catch {
    global linux_syscall
    array set args $argstring
    foreach var {sim mcm cpu thread sc} {eval set $var $args($var)}
    set cpu $mcm:$cpu:$thread

    set gpr3 [$sim cpu $cpu display gpr 3]
    set gpr4 [$sim cpu $cpu display gpr 4]
    set gpr5 [$sim cpu $cpu display gpr 5]
    set gpr6 [$sim cpu $cpu display gpr 6]
    set gpr7 [$sim cpu $cpu display gpr 7]
    set gpr8 [$sim cpu $cpu display gpr 8]

    switch $sc {
        1 { set info [format "exit(%d)" $gpr3] }
        2 { set info "fork()" }
        3 { set info [format "read(fd=%d,buf=0x%x,len=0x%x)" $gpr3 $gpr4 $gpr5] }
        4 { set info [format "write(fd=%d,buf=0x%x,len=0x%x)" $gpr3 $gpr4 $gpr5] }
        5 { set info [format "open(%s, ...)" [get_str $sim $cpu $gpr3]] }
        6 { set info [format "close(%d)" $gpr3] }
        7 { set info [format "waitpid(pid=%d,loc=0x%x,options=0x%x)" $gpr3 $gpr4 $gpr5] }
        8 { set info [format "link(oldname=%s,newname=%s)" [get_str $sim $cpu $gpr3] [get_str $sim $cpu $gpr4]] }
        10 { set info [format "unlink(pathname=%s)" [get_str $sim $cpu $gpr3]] }
        11 { set info [format "execve(%s, ...)" [get_str $sim $cpu $gpr3]] }
        12 { set info [format "chdir(%s)" [get_str $sim $cpu $gpr3]] }
        13 { set info "time(...)" }
        15 { set info [format "chmod(%s,0x%x)" [get_str $sim $cpu $gpr3] $gpr4] }
        19 { set info [format "lseek(fd=%d,offset=0x%x,origin=0x%x)" $gpr3 $gpr4 $gpr5] }
        20 { set info "getpid(...)" }
        21 { set info [format "mount(dev_name=%s,dir_name=%s,type=%s,flags=0x%x,data=0x%x)" \
                           [get_str $sim $cpu $gpr3] [get_str $sim $cpu $gpr4] \
                           [get_str $sim $cpu $gpr5] $gpr6 $gpr7] }
        23 { set info "setuid(...)" }
        24 { set info "getuid(...)" }
        27 { set info [format "alarm(seconds=%d)" $gpr3] }
        33 { set info [format "access(%s,mode=0x%x)" [get_str $sim $cpu $gpr3] $gpr4] }
        39 { set info [format "mkdir(%s,mode=0x%x)" [get_str $sim $cpu $gpr3] $gpr4] }
        41 { set info [format "dup(%d)" $gpr3] }
        42 { set info [format "pipe(filedes=%d)" $gpr3] }
        43 { set info "times(...)" }
        45 { set info [format "brk(0x%x)" $gpr3] }
        46 { set info "setgid(...)" }
        47 { set info "getgid(...)" }
        49 { set info "geteuid(...)" }
        50 { set info "getegid(...)" }
        54 { set info [format "ioctl(fd=%d,...)" $gpr3] }
        57 { set info [format "setpgid(pid=0x%x,pgid=0x%x)" $gpr3 $gpr4] }
        60 { set info [format "umask(mask=0x%x)" $gpr3] }
        63 { set info [format "dup2(%d,%d)" $gpr3 $gpr4] }
        64 { set info "getppid(...)" }
        65 { set info "getpgpr(...)" }
        66 { set info "setsid(...)" }
        78 { set info "gettimeofday(...)" }
        83 { set info [format "symlink(old=%s,new=%s)" [get_str $sim $cpu $gpr3] [get_str $sim $cpu $gpr4]] }
        85 { set info [format "readlink(%s,...)" [get_str $sim $cpu $gpr3]] }
        88 { set info [format "reboot(cmd=0x%x)" $gpr3] }
        90 { set info [format "mmap(addr=0x%x,len=0x%x,prot=0x%x,flags=0x%x,fd=%d,offset=0x%x)" $gpr3 $gpr4 $gpr5 $gpr6 $gpr7 $gpr8] }
        91 { set info [format "munmap(addr=0x%x,len=0x%x)" $gpr3 $gpr4] }
        102 { set info [format "socketcall(call=%d,args=0x%x)" $gpr3 $gpr4] }
        114 { set info [format "wait4(pid=%d,0x%x,0x%x,0x%x)" $gpr3 $gpr4 $gpr5 $gpr6] }
        116 { set info [format "sysinfo(sysinfo=0x%x)" $gpr3] }
        120 { set info [format "clone(fn=0x%x,child_stack=0x%x,flags=0x%x,...)" $gpr3 $gpr4 $gpr5] }
        122 { set info "uname(...)" }
        125 { set info [format "mprotect(addr=0x%x,len=0x%x,prot=0x%x)" $gpr3 $gpr4 $gpr5] }
        140 { set info [format "llseek(fd=%d,off_high=0x%x,off_low=0x%x,result=0x%x,whence=%d)" $gpr3 $gpr4 $gpr5 $gpr6 $gpr7] }
        142 { set info [format "select(fd=%d,...)" $gpr3] }
        162 { set info [format "nanosleep(rqtp=0x%x,rmtp=0x%x)" $gpr3 $gpr4] }
        173 { set info [format "rt_sigaction(%d)" $gpr3] }
        174 { set info [format "rt_sigprocmask(%d)" $gpr3] }
        181 { set info [format "chown(%s,uid=%d,gid=%d)" [get_str $sim $cpu $gpr3] $gpr4 $gpr5] }
        182 { set info "getcwd(...)" }
        190 { set info "ugetrlimit(....)" }
        192 { set info [format "mmap2(addr=0x%x,len=0x%x,prot=0x%x,flags=0x%x,fd=%d,off=0x%x)" $gpr3 $gpr4 $gpr5 $gpr6 $gpr7 $gpr8] }
        195 { set info [format "stat64(%s,...)" [get_str $sim $cpu $gpr3]] }
        196 { set info [format "lstat64(%s,statbuf=0x%x)" [get_str $sim $cpu $gpr3] $gpr4] }
        197 { set info [format "fstat64(fd=%d,..)" $gpr3] }
        202 { set info [format "getdests64(0x%x,0x%x,0x%x,0x%x)" $gpr3 $gpr4 $gpr5 $gpr6] }
        204 { set info [format "fcntl64(fd=%d,cmd=%d,...)" $gpr3 $gpr4] }
        205 { set info [format "madvise(addr=0x%x,len=0x%x,advise=%d)" $gpr3 $gpr4 $gpr5] }
        234 { set info "exit_group(...)" }
        252 { set info [format "statfs64(%s,...)" [get_str $sim $cpu $gpr3]] }
        default { set info [format "SC%d(0x%x, ...)" $sc $gpr3] }
    }

    if {[display regress_mode] == "off"} {
        set cycle [$sim display cycle]
        set inst_count [$sim display instruction_count]
    } else {
        set cycle "<c>"
        set inst_count "<i>"
    }

    puts "$cycle: $inst_count: $cpu : SysCall PC=[$sim cpu $cpu display spr pc] $info"

    } err]} { puts "[info level [info level]] failed: $err" }
}

proc syscall_trace::enable { { sim mysim } } {
    variable verbose
    variable max_syscall_number
    for {set i 1} {$i < $max_syscall_number} {incr i} {
        mysim trigger set syscall $i syscall_trace::syscall_trigger
    }
}

proc syscall_trace::disable { { sim mysim } } {
    variable verbose
    variable max_syscall_number
    for {set i 1} {$i < $max_syscall_number} {incr i} {
        mysim trigger clear syscall $i syscall_trace::syscall_trigger
    }
}
